import copy
from typing import Any, Iterator, Union, AsyncIterator, cast

from langchain.schema import Document

from langchain_rag.document_transformers.runnable_document_transformer import (
    RunnableGeneratorDocumentTransformer, to_async_iterator,
)


class CopyDocumentTransformer(RunnableGeneratorDocumentTransformer):
    def lazy_transform_documents(
            self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        yield from (copy.deepcopy(doc) for doc in documents)

    async def alazy_transform_documents(
            self, documents: Union[AsyncIterator[Document], Iterator[Document]],
            **kwargs: Any
    ) -> AsyncIterator[Document]:
        if isinstance(documents, AsyncIterator):
            async_documents = documents
        else:
            async_documents = to_async_iterator(documents)
        return async_documents



import copy
from typing import Any, Callable, Dict, List, Optional, cast, \
    Iterator

from langchain.chains import LLMChain
from langchain.output_parsers import PydanticOutputParser
from langchain.prompts import PromptTemplate
from langchain.pydantic_v1 import BaseModel
from langchain.schema import Document, BaseOutputParser
from langchain.schema.language_model import BaseLanguageModel

from langchain_rag.document_transformers.runnable_document_transformer import (
    RunnableGeneratorDocumentTransformer,
)


def _default_get_input(doc: Document) -> Dict[str, Any]:
    """Return the context chain input."""
    return {
        "context": doc.page_content,
    }


class SummarizeAndQuestions(BaseModel):
    summary: str
    """the document summary."""
    questions: List[str]
    """A list of questions"""


_default_parser: BaseOutputParser = PydanticOutputParser(
    pydantic_object=SummarizeAndQuestions)

_default_template = """
1. Given a text input, generate {nb_of_questions} questions from it in the same language. 
2. Sumarize a text input in the same language. 
Context:
```
{context}
```
{format_instructions}"""


def _get_default_chain_prompt() -> PromptTemplate:
    return PromptTemplate.from_template(
        template=_default_template,
        output_parser=_default_parser,
        partial_variables={
            "format_instructions": _default_parser.get_format_instructions()
        },
    )


class SummarizeAndQuestionsTransformer(RunnableGeneratorDocumentTransformer):
    """Generate questions and summarize for each Documents."""

    llm_chain: LLMChain
    get_input: Callable[[Document], dict] = _default_get_input
    nb_of_questions: int = 3

    def lazy_transform_documents(
            self,
            documents: Iterator[Document],
            **kwargs: Any
    ) -> Iterator[Document]:
        """Compress page content of raw documents."""
        _callbacks = kwargs.get("callbacks", None)
        for doc in documents:
            _input = {
                **self.get_input(doc),
                **{"nb_of_questions": self.nb_of_questions},
            }
            output = cast(
                SummarizeAndQuestions,
                self.llm_chain.predict(
                    callbacks=_callbacks,
                    **_input,
                ),
            )
            if not output:
                continue
            yield Document(page_content=output.summary, metadata=doc.metadata)
            for q in output.questions:
                metadata = copy.deepcopy(doc.metadata)
                metadata["transformer"] = self.__class__.__name__
                yield Document(page_content=q, metadata=metadata)

    @classmethod
    def from_llm(
            cls,
            llm: BaseLanguageModel,
            prompt: Optional[PromptTemplate] = None,
            get_input: Optional[Callable[[Document], dict]] = None,
            nb_of_questions: int = 3,
            llm_chain_kwargs: Optional[dict] = None,
    ) -> "SummarizeAndQuestionsTransformer":
        """Initialize from LLM."""
        _prompt = prompt if prompt is not None else _get_default_chain_prompt()
        _get_input = get_input if get_input is not None else _default_get_input
        llm_chain = LLMChain(
            llm=llm,
            prompt=_prompt,
            output_parser=_prompt.output_parser,
            **(llm_chain_kwargs or {}),
        )
        return cls(
            llm_chain=llm_chain, get_input=_get_input, nb_of_questions=nb_of_questions
        )



import asyncio
import copy
from collections.abc import AsyncIterator
from functools import partial
from typing import Any, Callable, Dict, Optional, Sequence, Iterator, Union, cast

from langchain.chains import LLMChain
from langchain.output_parsers import NumberedListOutputParser
from langchain.prompts import PromptTemplate
from langchain.schema import Document
from langchain.schema.language_model import BaseLanguageModel

from langchain_rag.document_transformers.runnable_document_transformer import (
    RunnableGeneratorDocumentTransformer, to_async_iterator,
)


def _default_get_input(doc: Document) -> Dict[str, Any]:
    """Return the context chain input."""
    return {
        "context": doc.page_content,
    }


_default_template = """
Sumarize a text input in the same language. 
Context:
```
{context}
```
"""
_default_format_instruction = NumberedListOutputParser()


def _get_default_chain_prompt() -> PromptTemplate:
    return PromptTemplate.from_template(
        template=_default_template,
    )


class SummarizeTransformer(RunnableGeneratorDocumentTransformer):
    """Generate questions for each Documents."""

    llm_chain: LLMChain
    get_input: Callable[[Document], dict] = _default_get_input

    """LLM wrapper to use for compressing documents."""

    """Callable for constructing the chain input from the query and a Document."""

    def lazy_transform_documents(
            self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Compress page content of raw documents."""
        _callbacks = kwargs.get("callbacks", None)
        for doc in documents:
            _input = self.get_input(doc)
            output = self.llm_chain.predict(callbacks=_callbacks, **_input)
            if not output:
                continue
            metadata = copy.deepcopy(doc.metadata)
            metadata["transformer"] = self.__class__.__name__
            yield Document(
                page_content="SUMMARY:\n" + str(output), metadata=metadata  # FIXME
            )

    def transform_documents(
            self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        return list(self.lazy_transform_documents(documents=iter(documents), **kwargs))

    async def alazy_transform_documents(  # type:ignore
            self, documents: Union[AsyncIterator[Document], Iterator[Document]],
            **kwargs: Any
    ) -> AsyncIterator[Document]:
        """Summarize the page content of raw documents asynchronously."""
        _callbacks = kwargs.get("callbacks", None)
        if isinstance(documents, AsyncIterator):
            async_documents = cast(AsyncIterator[Document], documents)
        else:
            async_documents = to_async_iterator(documents)

        outputs = await asyncio.gather(
            *[
                self.llm_chain.apredict(
                    **self.get_input(doc), callbacks=_callbacks
                )
                async for doc in async_documents
            ]
        )
        i = 0
        async for doc in async_documents:
            if not outputs[i]:
                continue
            yield Document(page_content=outputs[i], metadata=doc.metadata)
            i += 1

    async def atransform_documents(
            self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        """Asynchronously transform a list of documents.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns:
            A list of transformed Documents.
        """
        # FIXME: a tester. Lazy ?
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(self.transform_documents, **kwargs), documents
        )

    @classmethod
    def from_llm(
            cls,
            llm: BaseLanguageModel,
            prompt: Optional[PromptTemplate] = None,
            get_input: Optional[Callable[[Document], dict]] = None,
            llm_chain_kwargs: Optional[dict] = None,
    ) -> "SummarizeTransformer":
        """Initialize from LLM."""
        _prompt = prompt if prompt is not None else _get_default_chain_prompt()
        _get_input = get_input if get_input is not None else _default_get_input
        llm_chain = LLMChain(
            llm=llm,
            prompt=_prompt,
            **(llm_chain_kwargs or {}),
        )
        return cls(llm_chain=llm_chain, get_input=_get_input)



import itertools
import sys
from functools import partial
from typing import Any, Iterator, Sequence, Iterable, TypeVar, Tuple, \
    Union, AsyncIterator, Mapping

from langchain.schema import BaseDocumentTransformer, Document
from langchain.schema.runnable import RunnableGenerator, RunnableParallel, Runnable

from .runnable_document_transformer import RunnableGeneratorDocumentTransformer

if sys.version_info.major > 3 or sys.version_info.minor > 10:
    from itertools import batched  # type: ignore[attr-defined]
else:

    T = TypeVar('T')  # Only in python 3.12


    def batched(iterable: Iterable[T], n: int) -> Iterator[Tuple[T, ...]]:
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch

BATCH_SIZE = 16
_COMPATIBLE_RUNNABLE = True


def _transform_documents_generator(
        documents: Iterator[Document], *,
        transformers: Sequence[RunnableGeneratorDocumentTransformer]
) -> Iterator[Document]:
    Input = Union[AsyncIterator[Document], Iterator[Document]]
    steps = {
        f"transform_documents_{i}": transformer
        for i, transformer in enumerate(transformers)
    }
    # Implementation when all transformers are compatible with Runnable
    for batch in batched(documents, BATCH_SIZE):
        result = RunnableParallel[Input](steps=steps).invoke(batch)
        for chunk in result["steps"].values():
            yield chunk


class DocumentTransformers(RunnableGeneratorDocumentTransformer):
    """Document transformer that uses a list of Transformers."""

    class Config:
        arbitrary_types_allowed = True

    transformers: Sequence[BaseDocumentTransformer]
    # transformers: Sequence[RunnableGeneratorDocumentTransformer]  # FIXME: temporaire, pour tester
    """List of document transformer that are applied in parallel."""

    def lazy_transform_documents(
            self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Transform an interator of documents with the list of transformations.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns:
            An iterator oftransformed Documents.
        """
        if _COMPATIBLE_RUNNABLE:
            # Version stream
            result = RunnableGenerator[Document, Document](
                transform=partial(
                    _transform_documents_generator, transformers=self.transformers
                )
            ).transform(
                input=documents,
                **kwargs,
            )
            yield from (y for chunk in result for y in chunk)  # type: ignore
        else:
            # Implementation when all transformers are NOT compatible with Runnable
            # It's not compatible with lazy strategy. Load all documents and apply
            # all transformations.
            docs = list(documents)
            for transformer in self.transformers:
                for doc in transformer.transform_documents(documents=docs):
                    yield doc

    async def alazy_transform_documents(
            self, documents: Union[AsyncIterator[Document], Iterator[Document]],
            **kwargs: Any
    ) -> AsyncIterator[Document]:
        """Asynchronously transform an iterator of documents with a list of transformations.

        Args:
            documents: An iterator of Documents to be transformed.

        Returns:
            An interator of transformed Documents.
        """
        raise NotImplementedError("Not yet")



import asyncio
from abc import abstractmethod, ABC
from collections.abc import AsyncIterator
from typing import Any, Iterator, Optional, Sequence, Union, \
    Iterable, TypeVar

from langchain.schema import BaseDocumentTransformer, Document
from langchain.schema.runnable import RunnableConfig, RunnableSerializable

"""
    We propose an alternative way of making transformers compatible with LCEL.
    The first keeps the current protocol (RunnableDocumentTransformer).
    The second takes advantage of this to propose
    a lazy approach to transformations (RunnableGeneratorDocumentTransformer).
    It's better for the memory, pipeline, etc.
    
    Now, it's possible to create a pipeline of transformer like:
    Example:
    ..code - block:: python
    class UpperTransformer(RunnableGeneratorDocumentTransformer):
        def lazy_transform_documents(
                self,
                documents: Iterator[Document],
                **kwargs: Any
        ) -> Iterator[Document]:
            ...

        async def alazy_transform_documents(
                self,
                documents: Union[AsyncIterator[Document],Iterator[Document]],
                **kwargs: Any
        ) -> AsyncIterator[Document]:
            ...

    runnable = (UpperTransformer() | LowerTransformer())
    result = list(runnable.invoke(documents))
"""

T = TypeVar("T")


async def to_async_iterator(iterator: Iterable[T]) -> AsyncIterator[T]:
    """Convert an iterable to an async iterator."""
    for item in iterator:
        yield item


class RunnableGeneratorDocumentTransformer(
    RunnableSerializable[
        Union[AsyncIterator[Document], Iterator[Document]],  # input
        Union[AsyncIterator[Document], Iterator[Document]]  # output
    ],
    BaseDocumentTransformer,
    ABC
):
    """
    Runnable Document Transformer with lazy transformation.

    This class is a transition class for proposing lazy transformers, compatible with LCEL.
    Later, it can be integrated into BaseDocumentTransformer
    if you agree to add a lazy approach to transformations.
    All subclass of BaseDocumentTransformer must be updated to be compatible with this.
    """

    def transform_documents(
            self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        # Convert lazy to classical transformation
        return list(self.lazy_transform_documents(iter(documents), **kwargs))

    async def atransform_documents(
            self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        # Convert lazy to classical transformation
        return [
            doc
            async for doc in
                self.alazy_transform_documents(iter(documents), **kwargs)  # type:ignore
        ]

    @abstractmethod
    def lazy_transform_documents(
            self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Transform an interator of documents.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns:
            An iterator oftransformed Documents.
        """
        raise NotImplementedError()

    @abstractmethod
    async def alazy_transform_documents(
            self, documents: Union[AsyncIterator[Document], Iterator[Document]],
            **kwargs: Any
    ) -> AsyncIterator[Document]:
        """Asynchronously transform an iterator of documents.

        Args:
            documents: An iterator of Documents to be transformed.

        Returns:
            An interator of transformed Documents.
        """
        raise NotImplementedError()

    def invoke(
            self,
            input: Union[AsyncIterator[Document], Iterator[Document]],
            config: Optional[RunnableConfig] = None,
            **kwargs: Any,
    ) -> Union[AsyncIterator[Document], Iterator[Document]]:
        if isinstance(input, AsyncIterator):
            raise ValueError("Use ainvoke with async iterator")
        config = config or {}

        if hasattr(self, "lazy_transform_documents"):
            return self.lazy_transform_documents(input, **config)

        # Default implementation, without generator
        return iter(self.transform_documents(list(input), **config))

    async def ainvoke(
            self,
            input: Union[Iterable[Document], AsyncIterator[Document]],
            config: Optional[RunnableConfig] = None,
            **kwargs: Optional[Any],
    ) -> Union[AsyncIterator[Document], Iterator[Document]]:
        # # Default implementation, without generator
        config = config or {}
        return self.alazy_transform_documents(documents=input, **config)  # type:ignore


class RunnableDocumentTransformer(
    RunnableSerializable[Sequence[Document], Sequence[Document]],
    BaseDocumentTransformer,
    ABC
):
    """
    Runnable Document Transformer with lazy transformation.
    This class is a transition class for proposing lazy transformers, compatible with LCEL.
    Later, it can be integrated into BaseDocumentTransformer
    if you refuse to add a lazy approach to transformations.
    All subclass of BaseDocumentTransformer must be updated to be compatible with this.
    """
    """  # FIXME
    Now, it's possible to create a chain of transformations
    (only if the transformation is compatible with `RunnableDocumentTransformer`)
    """

    def invoke(
            self, input: Sequence[Document], config: Optional[RunnableConfig] = None
    ) -> Sequence[Document]:
        config = config or {}
        return self.transform_documents(
            input,
            **config,
        )

    async def ainvoke(
            self,
            input: Sequence[Document],
            config: Optional[RunnableConfig] = None,
            **kwargs: Optional[Any],
    ) -> Sequence[Document]:
        config = config or {}
        return await self.atransform_documents(input, **config)
